/***************************************************
-Version: R0.0
-Feature:  Hand Gesture(Left/Right Swipe)
          Slide Show On/Off(Button, default is on)
          Note for checking SD card
          fix Slide Show_off issue=>Show last bnbn
-Date:2019/01/05
 ****************************************************/
#include "Adafruit_GFX.h"      // Core graphics library
#include "Adafruit_ILI9340.h"  // Hardware-specific library
#include <SparkFun_APDS9960.h> // Gesture sensor library
#include <SPI.h>
#include <SD.h>
#include <Wire.h>
#include <string.h> 
// TFT display and SD card will share the hardware SPI interface.
// Hardware SPI pins are specific to the Arduino board type and
// cannot be remapped to alternate pins.  For Arduino Uno,
// Duemilanove, etc., pin 11 = MOSI, pin 12 = MISO, pin 13 = SCK.
#define TFT_RST 8
#define TFT_DC 9
#define TFT_CS 10
#define SD_CS 4
#define BUFFPIXEL 20
#define APDS9960_INT    2 // To be an APDS9960 interrupt pin
#define BUTTON_INT      3 // To be a button interrupt pin   

Adafruit_ILI9340 tft = Adafruit_ILI9340(TFT_CS, TFT_DC, TFT_RST); 
SparkFun_APDS9960 apds = SparkFun_APDS9960();

volatile int rotation=0;
volatile int tmp_rotation;
volatile boolean isr_flag = 0;
volatile boolean button_flag = 0; 
volatile boolean slide_show = 0;
volatile boolean displayed = 0;
File root;
uint8_t tmp_numberOfFiles = 0;
uint8_t numberOfFiles = 0;
int current_file_index = 0;
volatile char current_filename[15];
unsigned long previous_time,current_time;

void setup(void) {
  Serial.begin(9600);
  pinMode(APDS9960_INT, INPUT);  // Set interrupt pin as input
  pinMode(BUTTON_INT, INPUT);    // Set interrupt pin as input  
  pinMode(SD_CS, OUTPUT);
  pinMode(5, OUTPUT);          // Set Buzzer pin as output
  pinMode(6, OUTPUT);          // Set Index LED pin as output
  digitalWrite(SD_CS, HIGH);
  digitalWrite(5, LOW);
  digitalWrite(6, LOW);
  tft.begin();
  tft.fillScreen(ILI9340_WHITE);
  /****** Check SD card ********/
  Serial.print(F("Initializing SD card..."));
  if (!SD.begin(SD_CS)) {
    Serial.println(F("failed!"));
    rotation=1;
    tft.setRotation(rotation);
    tft.setCursor(30, 100);
    tft.setTextColor(ILI9340_BLACK);  
    tft.setTextSize(2);
    tft.println("Please check SD card");
    tft.println("& Power reset");
    rotation=0;
    delay(10000);
    return;
  }
  Serial.println(F("SD_OK!"));
  /****** Show logo.bmp for 5 Sec.********/
  bmpDraw("logo.bmp", 0, 0);    
  delay(5000);
  tft.fillScreen(ILI9340_WHITE);
  /******* Read SD Files & Rewind dircetory *********/
  root = SD.open("/");
  root.rewindDirectory();
  printDirectory(root, 0,current_file_index);
  numberOfFiles=tmp_numberOfFiles;
  root.close();
  /****** Initialize APDS-9960 (configure I2C and initial values)****/
  attachInterrupt(0, interruptRoutine, FALLING);
  if ( apds.init() ) {
    Serial.println(F("APDS-9960 initialization complete"));
  } else {
    Serial.println(F("Something went wrong during APDS-9960 init!"));
  } 
  // Start running the APDS-9960 gesture sensor engine
  if ( apds.enableGestureSensor(true) ) {
    Serial.println(F("Gesture sensor is now running"));
    swipeText();
    previous_time= millis();
  } else {
    Serial.println(F("Something went wrong during gesture sensor init!"));
  }
  /****** Button Trigger****/
  attachInterrupt(1,button_interrupt, RISING);
  /*******Slide Show=On (Default)****/
  slide_show=1;
}

void loop() {
  
        if(slide_show==1){
            current_time=millis();
     
            /***** Slide Show >=7 Sec.*****/
            if(current_time-previous_time>=7000){  
              //Serial.println("Slide Show");
              current_file_index++;
              if(current_file_index>numberOfFiles)
              {
              current_file_index=1;
               }
             /****** Show Current Index Picture*****/
              root = SD.open("/");
              root.rewindDirectory();
              printDirectory(root,0,current_file_index);
              root.close();
              bmpDraw(current_filename, 0, 0);
             /****** Show Current Index Picture*****/
              previous_time=millis();
            }
          }
          else{
               current_time=millis();
              
               if(displayed==1){
               
                if(current_time-previous_time>=1500){ 
                /****** Show Current Index Picture*****/
                  root = SD.open("/");
                  root.rewindDirectory();
                  printDirectory(root,0,current_file_index);
                  root.close();
                  bmpDraw(current_filename, 0, 0);
                  displayed=!displayed;
                  /****** Show Current Index Picture*****/
                  previous_time=millis();
                  
                  }
               }
            }
        if( isr_flag == 1 ) {
              detachInterrupt(0);
              handleGesture();
              displayed=1;
              isr_flag = 0;
              attachInterrupt(0, interruptRoutine, FALLING);
              previous_time=millis();
          }
        if( button_flag == 1 ) {
              Serial.println(F("Button"));
              slide_show=!slide_show;
              Serial.println(slide_show);
              previous_time=millis();
              //displayed=1;
              button_flag = 0;
              attachInterrupt(1, button_interrupt, RISING);
         }
}

/**** Filter .bmp files in SD card***/
int getExtension(char *fileName){   
   int len = strlen(fileName);   
   while( fileName[len]!='.' && len>0 ){ len--; }  
   if(fileName[len]=='.' && fileName[len+1]=='B' && fileName[len+2]=='M' && fileName[len+3]=='P'){  
          strcpy(current_filename, fileName);
          //Serial.print( current_filename);
          tmp_numberOfFiles=tmp_numberOfFiles+1;  
       }  
        return tmp_numberOfFiles;
  } 
 
/****** Print out files in SD card  *****/
void printDirectory(File dir, int numTabs, int current_file_index)
{
  tmp_numberOfFiles=0;
  while (true)
  {
    File entry = dir.openNextFile();
    if (! entry)
    {
      if (numTabs == 0)
        //dir.rewindDirectory();
        Serial.println("** Done **");
      return;
    }
    for (uint8_t i = 0; i < numTabs; i++)
    //Serial.print('\t');
      entry.name();
    //Serial.println(entry.name());
   /* 
    if (entry.isDirectory())
    {
      //Serial.println("/");
      printDirectory(entry, numTabs + 1, 0);
    }
    else
    {
    */
      getExtension(entry.name());   // To filter out BMP files 
      if(current_file_index>0 && tmp_numberOfFiles==current_file_index)
      { 
          entry.close();
          return;
        }
    /* } */  
    entry.close();
  }
}
/******* Swipe Text Display ********/
void swipeText() {
  tmp_rotation=rotation;
  tft.fillScreen(ILI9340_BLACK);
  rotation=1;
  tft.setRotation(rotation);
  tft.setCursor(30, 100);
  tft.setTextColor(ILI9340_WHITE);  
  tft.setTextSize(2);
  tft.println("Swipe Left or Right");
  rotation=tmp_rotation;
}

/******* Set APD9660 interrupt flag = 1********/
void interruptRoutine() {
  isr_flag = 1;
}

/******* Set button interrupt flag = 1********/
void button_interrupt() {
 static unsigned long last_interrupt_time = 0;
 unsigned long interrupt_time = millis();
 if (interrupt_time - last_interrupt_time > 300)
 {
  button_flag = 1;
 }
  last_interrupt_time = interrupt_time;
}

/******* Gesture Function ********/
void handleGesture() {
    if ( apds.isGestureAvailable() ) {
    switch ( apds.readGesture() ) {
      case DIR_UP:
       // Serial.println("UP");
        /*
        rotation++;
         if(rotation>3)
            {
             rotation=0;
             }
         */
        /****** Show Current Index Picture*****/
        /*
        root = SD.open("/");
        root.rewindDirectory();
        printDirectory(root,0,current_file_index);
        root.close();
        bmpDraw(current_filename, 0, 0);
        */
        /****** Show Current Index Picture*****/
        //Serial.println(rotation);
        swipeText();
        break;
      case DIR_DOWN:
        //Serial.println("DOWN");
        /*
         rotation--;
         if(rotation<0)
            {
             rotation=3;
             }
         */
        /****** Show Current Index Picture*****/
        /*
        root = SD.open("/");
        root.rewindDirectory();
        printDirectory(root,0,current_file_index);
        root.close();
        bmpDraw(current_filename, 0, 0);
        */
        /****** Show Current Index Picture*****/ 
        //Serial.println(rotation);
        swipeText();
        break;
      case DIR_LEFT:
        //Serial.println(F("LEFT"));
        current_file_index--;
        if(current_file_index<=0)
         {
          current_file_index=numberOfFiles;
          }
          /****** Show Current Index Picture*****/
          root = SD.open("/");
          root.rewindDirectory();
          printDirectory(root,0,current_file_index);
          root.close();
          bmpDraw(current_filename, 0, 0);
          /****** Show Current Index Picture*****/
          break;
      case DIR_RIGHT:
          //Serial.println(F("RIGHT"));
          current_file_index++;
          if(current_file_index>numberOfFiles)
            {
             current_file_index=1;
             }
          /****** Show Current Index Picture*****/
          root = SD.open("/");
          root.rewindDirectory();
          printDirectory(root,0,current_file_index);
          root.close();
          bmpDraw(current_filename, 0, 0);
          /****** Show Current Index Picture*****/
          break;
      case DIR_NEAR:
        //Serial.println("NEAR");
        swipeText();
        break;
      case DIR_FAR:
        //Serial.println("FAR");
        swipeText();
        break;
      default:
        //Serial.println("NONE");
        swipeText();
    }
  }
}
/******* Open BMP files on SD card ******/
// This function opens a Windows Bitmap (BMP) file and
// displays it at the given coordinates.  It's sped up
// by reading many pixels worth of data at a time
// (rather than pixel by pixel).  Increasing the buffer
// size takes more of the Arduino's precious RAM but
// makes loading a little faster.  20 pixels seems a
// good balance.

void bmpDraw(char *filename, uint16_t x, uint16_t y) {

  File     bmpFile;
  int      bmpWidth, bmpHeight;   // W+H in pixels
  uint8_t  bmpDepth;              // Bit depth (currently must be 24)
  uint32_t bmpImageoffset;        // Start of image data in file
  uint32_t rowSize;               // Not always = bmpWidth; may have padding
  uint8_t  sdbuffer[3*BUFFPIXEL]; // pixel buffer (R+G+B per pixel)
  uint8_t  buffidx = sizeof(sdbuffer); // Current position in sdbuffer
  boolean  goodBmp = false;       // Set to true on valid header parse
  boolean  flip    = true;        // BMP is stored bottom-to-top
  int      w, h, row, col;
  uint8_t  r, g, b;
  uint32_t pos = 0, startTime = millis();

  tft.setRotation(rotation);  // Can roate BMP's direction
  tft.fillScreen(ILI9340_BLACK);
  if((x >= tft.width()) || (y >= tft.height())) return;
   
  // Open requested file on SD card
  if ((bmpFile = SD.open(filename)) == NULL) {
    Serial.print( SD.open(filename));
    Serial.print("File not found");
    return;
  }

  // Parse BMP header
  if(read16(bmpFile) == 0x4D42) { // BMP signature
   // Serial.print("File size: "); Serial.println(read32(bmpFile));
   read32(bmpFile);
    (void)read32(bmpFile); // Read & ignore creator bytes
    bmpImageoffset = read32(bmpFile); // Start of image data
   // Serial.print("Image Offset: "); Serial.println(bmpImageoffset, DEC);
    // Read DIB header
   // Serial.print("Header size: "); Serial.println(read32(bmpFile));
    read32(bmpFile);
    bmpWidth  = read32(bmpFile);
    bmpHeight = read32(bmpFile);
    if(read16(bmpFile) == 1) { // # planes -- must be '1'
      bmpDepth = read16(bmpFile); // bits per pixel
     // Serial.print("Bit Depth: "); Serial.println(bmpDepth);
      if((bmpDepth == 24) && (read32(bmpFile) == 0)) { // 0 = uncompressed

        goodBmp = true; // Supported BMP format -- proceed!
       // Serial.print("Image size: ");
       // Serial.print(bmpWidth);
       // Serial.print('x');
       // Serial.println(bmpHeight);

        // BMP rows are padded (if needed) to 4-byte boundary
        rowSize = (bmpWidth * 3 + 3) & ~3;

        // If bmpHeight is negative, image is in top-down order.
        // This is not canon but has been observed in the wild.
        if(bmpHeight < 0) {
          bmpHeight = -bmpHeight;
          flip      = false;
        }

        // Crop area to be loaded
        w = bmpWidth;
        h = bmpHeight;
        if((x+w-1) >= tft.width())  w = tft.width()  - x;
        if((y+h-1) >= tft.height()) h = tft.height() - y;

        // Set TFT address window to clipped image bounds
        tft.setAddrWindow(x, y, x+w-1, y+h-1);

        for (row=0; row<h; row++) { // For each scanline...

          // Seek to start of scan line.  It might seem labor-
          // intensive to be doing this on every line, but this
          // method covers a lot of gritty details like cropping
          // and scanline padding.  Also, the seek only takes
          // place if the file position actually needs to change
          // (avoids a lot of cluster math in SD library).
          if(flip) // Bitmap is stored bottom-to-top order (normal BMP)
            pos = bmpImageoffset + (bmpHeight - 1 - row) * rowSize;
          else     // Bitmap is stored top-to-bottom
            pos = bmpImageoffset + row * rowSize;
          if(bmpFile.position() != pos) { // Need seek?
            bmpFile.seek(pos);
            buffidx = sizeof(sdbuffer); // Force buffer reload
          }

          for (col=0; col<w; col++) { // For each pixel...
            // Time to read more pixel data?
            if (buffidx >= sizeof(sdbuffer)) { // Indeed
              bmpFile.read(sdbuffer, sizeof(sdbuffer));
              buffidx = 0; // Set index to beginning
            }

            // Convert pixel from BMP to TFT format, push to display
            b = sdbuffer[buffidx++];
            g = sdbuffer[buffidx++];
            r = sdbuffer[buffidx++];
            tft.pushColor(tft.Color565(r,g,b));
          } // end pixel
        } // end scanline
     //   Serial.print("Loaded in ");
     //   Serial.print(millis() - startTime);
     //   Serial.println(" ms");
      } // end goodBmp
    }
  }

  bmpFile.close();
  if(!goodBmp) Serial.println(F("BMP format not recognized."));
}

// These read 16- and 32-bit types from the SD card file.
// BMP data is stored little-endian, Arduino is little-endian too.
// May need to reverse subscript order if porting elsewhere.

uint16_t read16(File & f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File & f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}
